/* Utilities: long long, clock */

#include "mpsys.h"

/* ----- BIT TWIDDLING */

/* lowest bits of an integer */
int int_lowest (int in, int bits)
{
	assert (bits <= sizeof(int)*8);
	return in & ((1<<bits)-1);
}

int in_range (int in, int min, int max)
{
	return (in >= min) && (in <= max);
} 

int in_pos_range (int in, int max)
{
	return (in >= 0) && (in <= max);
}

int is_boolean (int in)
{
	return (in == 0) || (in == 1); 
}

/* ----- buffer */

int buffer_invariant (buffer *b)
{
	return b && (b->memory) && (b->len>=0);
}

void buffer_reset (buffer *b, unsigned char *ptr, int len)
{
	assert (ptr);
	assert (len>=0);
	
	b->memory = ptr;
	b->len = len;
	
	assert (buffer_invariant (b));
}


/* hex dump */
void buffer_dump (buffer *b_in, FILE *f)
{
	buffer_iter b;
	
	assert (buffer_invariant (b_in));
	
	buffer_iter_reset (&b, b_in);
	buffer_iter_dump (&b, f);

}

/* ----- buffer_iter */

/* invariant for buffer_iter */
int buffer_iter_invariant (buffer_iter *b)
{
	return (b)
		&& ((b->len == 0) || ((b->memory) && (b->len > 0))) 
		&& in_pos_range (b->byte_pos, b->len) /* in buffer, allowing 'after' */ 
		&& in_pos_range (b->bit_pos, 7) /* bit pos in byte */
		&& ((b->bit_pos == 0) || (b->byte_pos < b->len)) /* (len,0) is the only 'after' pos */
		;
}

/* buffer iter of zero lenght */
void buffer_iter_zero (buffer_iter *b)
{
	b->memory = 0;
	b->len = 0;
	b->byte_pos = 0;
	b->bit_pos = 0;
	assert (buffer_iter_invariant (b));
}

/* create buffer reset */
void buffer_iter_reset (buffer_iter *b, buffer *in)
{
	/*require*/
	assert (b);
	
	buffer_iter_zero (b);
	b->memory = in->memory;
	b->len = in->len;
	
	/*ensure*/
	assert (b->byte_pos == 0); /* at start */
	assert (b->bit_pos == 0);
	assert (buffer_iter_invariant (b));
}

/* skip bytes */
void buffer_iter_skip_bytes (buffer_iter *b, int nbytes)
{
	assert (buffer_iter_invariant (b));
	/* require */
	assert (nbytes >= 0);
	assert (nbytes + b->byte_pos <= b->len);
	
	b->byte_pos = b->byte_pos + nbytes;
	
	assert (buffer_iter_invariant (b)); /* includes alignment/after overflow */
}

/* skip bits */
void buffer_iter_skip_bits (buffer_iter *b, int nbits)
{
	assert (buffer_iter_invariant (b));
	/* require */
	assert (nbits >= 0);
	assert (nbits <= buffer_iter_available_bits (b));
	
	buffer_iter_skip_bytes (b, nbits >> 3); /* bytes */
	buffer_iter_int_forth (b, nbits & 7); /* remaining bits */

	assert (buffer_iter_invariant (b)); /* includes overflow */
}

/* buffer iter truncate so that nbytes are available*/
void buffer_iter_truncate_bytes (buffer_iter *b, int nbytes)
{
	assert (buffer_iter_invariant (b));
	assert (nbytes >= 0);
	assert (nbytes <= buffer_iter_available_bytes (b));

	b->len = b->byte_pos + nbytes;
	if (b->bit_pos != 0)
	{
		(b->len)++;
	}
	assert (buffer_iter_invariant (b));
	assert (buffer_iter_available_bytes (b) ==  nbytes);
}

/* remaining bits in iterator */
int buffer_iter_available_bits (buffer_iter *b)
{
	int result = (b->len - b->byte_pos - 1) * 8 + (8 - b->bit_pos);
	
	assert (buffer_iter_invariant (b));
	assert ((b->len > b->byte_pos) || (result == 0)); /* after pos */
	return result;
}

/* remaining full bytes (may be a few more bits) */
int buffer_iter_available_bytes (buffer_iter *b)
{
	int result = (buffer_iter_available_bits (b) >> 3);
	
	assert (buffer_iter_invariant (b));
	assert (buffer_iter_available_bits (b) - (result * 8) < 8);
	assert (buffer_iter_available_bits (b) - (result * 8) >= 0);
	return result;
}

/* get int from buffer, and go forth in buffer */
int buffer_iter_int_forth (buffer_iter *b, int bitlen)
{
	int result = 0;

	assert (buffer_iter_invariant (b));
	/*require*/
	assert (bitlen >= 0);
	assert (bitlen <= (sizeof(int)*8));
	assert (bitlen <= buffer_iter_available_bits (b));
	
	if (((bitlen & 7) == 0) && (b->bit_pos == 0)) 
	{ /* when aligned */
		int i;
		
		for(i = 0; i < (bitlen>>3); i++)
		{
			result <<= 8;
			result |= b->memory[b->byte_pos];
			(b->byte_pos)++;
		}
	}
	else
	{ /* generic case */
		int avail_bits = (8 - b->bit_pos);
		int bitlen_left = bitlen;
		
		while (bitlen_left >= avail_bits)
		{ /* get all remaining bits in current byte */
			result <<= avail_bits;
			result |= int_lowest(b->memory[b->byte_pos], avail_bits);
			bitlen_left -= avail_bits;
			
			(b->byte_pos)++; /* forth */
			b->bit_pos = 0;
			avail_bits = 8;
		}
		
		assert (bitlen_left < avail_bits);
		
		if (bitlen_left>0)
		{
			result <<= bitlen_left;
			result |= int_lowest (b->memory[b->byte_pos] >> (avail_bits - bitlen_left), bitlen_left);
			b->bit_pos += bitlen_left;
		}
	}
	
	assert (buffer_iter_invariant(b)); /* also tests for overflow */
	return result;
}

/* mem copy between two iterated blocks (requires byte aligned!) */
void buffer_iter_memory_copy (buffer_iter *dest, buffer_iter *src, int nbytes)
{
	assert (buffer_iter_invariant (dest));
	assert (buffer_iter_invariant (src));
	/* require */
	assert (nbytes >= 0);
	assert (nbytes <= buffer_iter_available_bytes (src));
	assert (nbytes <= buffer_iter_available_bytes (dest));
	assert (dest->bit_pos == 0); /* aligned */
	assert (src->bit_pos == 0); /* aligned */
	
	memcpy (dest->memory + dest->byte_pos, src->memory + src->byte_pos, nbytes);
	
	assert (buffer_iter_invariant (dest));
	assert (buffer_iter_invariant (src));
}

/* convenience alias for skip_bytes s */
char *buffer_iter_skip_bytes_or_error (buffer_iter *b, int nbytes, char *error)
{
	assert (buffer_iter_invariant (b));
	assert (nbytes >= 0);

	if (buffer_iter_available_bytes (b) < nbytes)
	{
		return error;
	}
	else
	{
		buffer_iter_skip_bytes (b, nbytes);
		return 0;
	}
}

/* convenience alias for available_bytes */
char *buffer_iter_available_bytes_or_error (buffer_iter *b, int nbytes, char *error)
{
	assert (buffer_iter_invariant (b));
	assert (nbytes >= 0);

	if (buffer_iter_available_bytes (b) < nbytes)
	{
		return error;
	}
	else
	{
		return 0;
	}
}

/* pointer to memory at iterator's current position */
unsigned char *buffer_iter_memory_at (buffer_iter *b)
{
	assert (buffer_iter_invariant (b));
	return b->memory + b->byte_pos;
}

/* buffer at iterator's current position */
buffer buffer_iter_buffer_at (buffer_iter *b)
{
	buffer result;
	assert (buffer_iter_invariant (b));
	assert (b->bit_pos == 0); /* aligned */
	
	buffer_reset (&result, buffer_iter_memory_at (b), buffer_iter_available_bytes (b));
	return result;
}

/* buffer from start up to current position (exclusive) */
buffer buffer_iter_buffer_before (buffer_iter *b)
{
	buffer result;
	assert (buffer_iter_invariant (b));
	assert (b->bit_pos == 0); /* aligned */
	
	buffer_reset (&result, b->memory, b->byte_pos);
	
	assert (result.memory == b->memory);
	assert (result.len == b->byte_pos);
	return result;
}

/* new buffer_iter from start up to current position (exclusive) */
buffer_iter buffer_iter_buffer_iter_before (buffer_iter *b)
{
	buffer buffer;
	buffer_iter result;

	assert (buffer_iter_invariant (b));
	assert (b->bit_pos == 0); /* aligned */

	buffer = buffer_iter_buffer_before (b);
	buffer_iter_reset (&result, &buffer);
	
	assert (buffer_iter_available_bytes (&result) == b->byte_pos);
	return result;
}

/* hex dump */
void buffer_iter_dump (buffer_iter *b_in, FILE *f)
{
	int i = 1;
	buffer_iter b = *b_in;
	
	assert (buffer_iter_invariant (b_in));
	
	while (buffer_iter_available_bytes (&b) > 0)
	{
		fprintf (f, "%02x ", buffer_iter_int_forth (&b,8));
		if ((i % 32)==0)
		{ 
			fprintf (f, "\n"); 
		}
		i++;
	}
	fprintf (f, "\n"); 
}

/* print iter_in from current position to end, assuming text with XML escaping */
void buffer_iter_xml_text_dump (buffer_iter *iter_in, FILE *f)
{
	buffer_iter iter = *iter_in;
	while (buffer_iter_available_bytes (&iter) > 0)
	{
		unsigned char ch = (char) buffer_iter_int_forth (&iter, 8);
		if (ch == '<')
		{
			fprintf (f, "&lt;");
		}
		else if (ch == '&')
		{
			fprintf (f, "&amp;");
		}
		else if (ch == '>')
		{
			fprintf (f, "&gt;");
		}
		else if (ch == '\'')
		{
			fprintf (f, "&apos;");
		}
		else if (ch == '\"')
		{
			fprintf (f, "&quot;");
		}
		else if (ch > 0x7f)
		{
			fprintf (f, "&#%d;", (int) ch);
		}
		else if (ch >= 0x20)
		{
			fprintf (f, "%c", ch);
		}
		else
		{
			/* control char */
			/*fprintf (f, "<!-- control-char %d -->", (int) ch);*/
		}
	}
}

/* BCD */
int buffer_iter_bcd_forth (buffer_iter *b, int nbits)
{
	int i = nbits / 4;
	int result = 0;
	assert (buffer_iter_invariant (b));
	assert ((nbits % 4) == 0);
	assert (buffer_iter_available_bits (b) >= nbits);
	
	while (i > 0)
	{
		result = (10 * result) + buffer_iter_int_forth (b, 4);
		i--;
	}
	return result;
}

/* ----- long long */

mpsys_long mpsys_long_zero ()
{
	mpsys_long result;
	result.high = 0;
	result.low = 0;
	
	assert (result.high == 0 && result.low == 0);
	return result;
}

void mpsys_long_print (mpsys_long *l, FILE *f)
{
	assert (f);
	assert (l);
	assert (MPSYS_LONG_BASE == 24);
	fprintf (f, "%x%06x", l->high, l->low);
}

/* ----- Clock reference */

void program_clock_reference_print (program_clock_reference *c, FILE *f)
{
	assert (c);
	assert (MPSYS_LONG_BASE == 24);
	fprintf (f, "%03x%06x/%x ", c->base.high, c->base.low, c->extension);
}
